/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others
Copyright (C) 2007-2008 Kristian Duske
Copyright (C) 2010 Ozkan Sezer and Steven Atkinson

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// gl_vidnt.c -- NT GL vid component

#include "quakedef.h"
#include "resource.h"

void R_CreateRTTTextures (int width, int height);

// this should already be handled by the SDL GL include but just in case...
#ifndef APIENTRY
#define APIENTRY
#endif

gl_brokendrivers_t gl_brokendrivers;

#define MAX_MODE_LIST	600 // johnfitz -- was 30
#define VID_ROW_SIZE	3
#define WARP_WIDTH		320
#define WARP_HEIGHT		200
#define MAXWIDTH		10000
#define MAXHEIGHT		10000
#define BASEWIDTH		320
#define BASEHEIGHT		200

#define SDL_DEFAULT_FLAGS	(SDL_OPENGL | SDL_DOUBLEBUF)

typedef struct
{
	modestate_t	type;
	int			width;
	int			height;
	int			modenum;
	int			dib;
	int			fullscreen;
	int			bpp;
	// 	int			refreshrate; // johnfitz
	int			halfscreen;
	char		modedesc[17];
	qboolean	warn;
} vmode_t;

const char *gl_vendor = NULL;
const char *gl_renderer = NULL;
const char *gl_version = NULL;
const char *gl_extensions = NULL;

qboolean		DDActive;
qboolean		scr_skipupdate;

static vmode_t	modelist[MAX_MODE_LIST];
static int		nummodes;
static vmode_t	badmode;

static qboolean	vid_initialized = false;
static qboolean	windowed, leavecurrentmode;
static qboolean vid_canalttab = false;
extern qboolean	mouseactive;  // from in_win.c

SDL_Surface	*draw_context;

int			vid_modenum = NO_MODE;
int			vid_realmode;
int			vid_default = MODE_WINDOWED;
unsigned char	vid_curpal[256*3];

viddef_t	vid;				// global video state

// unsigned short	d_8to16table[256]; // johnfitz -- never used
// unsigned char		d_15to8table[65536]; // johnfitz -- never used

modestate_t	modestate = MODE_WINDOWED;

void VID_Menu_Init (void); // johnfitz
void VID_Menu_f (void); // johnfitz
void VID_MenuDraw (void);
void VID_MenuKey (int key);

char *VID_GetModeDescription (int mode);
void ClearAllStates (void);
void VID_UpdateWindowStatus (void);
void GL_Init (void);


PFNGLSECONDARYCOLORPOINTERPROC qglSecondaryColorPointer = NULL;
PFNGLMULTITEXCOORD2FARBPROC qglMultiTexCoord2fARB = NULL; // johnfitz
PFNGLACTIVETEXTUREARBPROC qglActiveTexture = NULL; // johnfitz
PFNGLACTIVETEXTUREARBPROC qglClientActiveTexture = NULL;	// mh

PFNGLGENQUERIESPROC qglGenQueries = NULL;
PFNGLDELETEQUERIESPROC qglDeleteQueries = NULL;
PFNGLISQUERYPROC qglIsQuery = NULL;
PFNGLBEGINQUERYPROC qglBeginQuery = NULL;
PFNGLENDQUERYPROC qglEndQuery = NULL;
PFNGLGETQUERYIVPROC qglGetQueryiv = NULL;
PFNGLGETQUERYOBJECTIVPROC qglGetQueryObjectiv = NULL;
PFNGLGETQUERYOBJECTUIVPROC qglGetQueryObjectuiv = NULL;
PFNGLBINDBUFFERPROC qglBindBuffer = NULL;
PFNGLDELETEBUFFERSPROC qglDeleteBuffers = NULL;
PFNGLGENBUFFERSPROC qglGenBuffers = NULL;
PFNGLISBUFFERPROC qglIsBuffer = NULL;
PFNGLBUFFERDATAPROC qglBufferData = NULL;
PFNGLBUFFERSUBDATAPROC qglBufferSubData = NULL;
PFNGLGETBUFFERSUBDATAPROC qglGetBufferSubData = NULL;
PFNGLMAPBUFFERPROC qglMapBuffer = NULL;
PFNGLUNMAPBUFFERPROC qglUnmapBuffer = NULL;
PFNGLGETBUFFERPARAMETERIVPROC qglGetBufferParameteriv = NULL;
PFNGLGETBUFFERPOINTERVPROC qglGetBufferPointerv = NULL;

PFNGLBLENDEQUATIONPROC qglBlendEquation = NULL;

// vertex/fragment program
PFNGLBINDPROGRAMARBPROC qglBindProgramARB = NULL;
PFNGLDELETEPROGRAMSARBPROC qglDeleteProgramsARB = NULL;
PFNGLGENPROGRAMSARBPROC qglGenProgramsARB = NULL;
PFNGLGETPROGRAMENVPARAMETERDVARBPROC qglGetProgramEnvParameterdvARB = NULL;
PFNGLGETPROGRAMENVPARAMETERFVARBPROC qglGetProgramEnvParameterfvARB = NULL;
PFNGLGETPROGRAMLOCALPARAMETERDVARBPROC qglGetProgramLocalParameterdvARB = NULL;
PFNGLGETPROGRAMLOCALPARAMETERFVARBPROC qglGetProgramLocalParameterfvARB = NULL;
PFNGLGETPROGRAMSTRINGARBPROC qglGetProgramStringARB = NULL;
PFNGLGETPROGRAMIVARBPROC qglGetProgramivARB = NULL;
PFNGLISPROGRAMARBPROC qglIsProgramARB = NULL;
PFNGLPROGRAMENVPARAMETER4DARBPROC qglProgramEnvParameter4dARB = NULL;
PFNGLPROGRAMENVPARAMETER4DVARBPROC qglProgramEnvParameter4dvARB = NULL;
PFNGLPROGRAMENVPARAMETER4FARBPROC qglProgramEnvParameter4fARB = NULL;
PFNGLPROGRAMENVPARAMETER4FVARBPROC qglProgramEnvParameter4fvARB = NULL;
PFNGLPROGRAMLOCALPARAMETER4DARBPROC qglProgramLocalParameter4dARB = NULL;
PFNGLPROGRAMLOCALPARAMETER4DVARBPROC qglProgramLocalParameter4dvARB = NULL;
PFNGLPROGRAMLOCALPARAMETER4FARBPROC qglProgramLocalParameter4fARB = NULL;
PFNGLPROGRAMLOCALPARAMETER4FVARBPROC qglProgramLocalParameter4fvARB = NULL;
PFNGLPROGRAMSTRINGARBPROC qglProgramStringARB = NULL;

PFNGLVERTEXATTRIBPOINTERARBPROC qglVertexAttribPointerARB = NULL;
PFNGLENABLEVERTEXATTRIBARRAYARBPROC qglEnableVertexAttribArrayARB = NULL;
PFNGLDISABLEVERTEXATTRIBARRAYARBPROC qglDisableVertexAttribArrayARB = NULL;

PFNGLDRAWRANGEELEMENTSPROC qglDrawRangeElements = NULL;
PFNGLLOCKARRAYSEXTPROC qglLockArrays = NULL;
PFNGLUNLOCKARRAYSEXTPROC qglUnlockArrays = NULL;
PFNGLMULTIDRAWARRAYSPROC qglMultiDrawArrays = NULL;

qboolean gl_arb_texture_compression = false;
qboolean gl_ext_compiled_vertex_array = false;

qboolean gl_swap_control = false; // johnfitz
qboolean gl_ext_texture_filter_anisotropic = false; // johnfitz
float gl_max_anisotropy; // johnfitz

int			gl_stencilbits; // johnfitz

qboolean vid_locked = false; // johnfitz
qboolean vid_changed = false;

void GL_SetupState (void); // johnfitz
void R_DeleteShaders (void);

// ====================================

// johnfitz -- new cvars
cvar_t		vid_fullscreen = {"vid_fullscreen", "0", true};	// QuakeSpasm, was "1"
cvar_t		vid_width = {"vid_width", "800", true};		// QuakeSpasm, was 640
cvar_t		vid_height = {"vid_height", "600", true};	// QuakeSpasm, was 480
cvar_t		vid_bpp = {"vid_bpp", "32", true};			// mh - was 16
cvar_t		vid_refreshrate = {"vid_refreshrate", "60", true};
cvar_t		vid_vsync = {"vid_vsync", "0", true};
// johnfitz

cvar_t		_windowed_mouse = {"_windowed_mouse", "1", true};
cvar_t		vid_gamma = {"gamma", "1", true}; // johnfitz -- moved here from view.c
cvar_t		vid_contrast = {"contrast", "1", true};


void VID_PrintGLExtensions (void)
{
	int i, j, pos;
	char *currext = (char *) scratchbuf;

	Con_SafePrintf ("GL_EXTENSIONS:\n");

	// prevent Con_Printf buffer overflow when printing extensions
	for (i = 0, pos = 0; ; i++)
	{
		// no more extensions
		if (!gl_extensions[i]) break;

		if (gl_extensions[i] == ' ')
		{
			for (j = 0; ; j++)
			{
				if (gl_extensions[pos + j] == ' ')
				{
					currext[j] = 0;
					break;
				}
				else currext[j] = gl_extensions[pos + j];
			}

			Con_SafePrintf ("   %s\n", currext);
			pos = i + 1;
		}
	}
}


// ==========================================================================
//
//  HARDWARE GAMMA -- johnfitz
//
// ==========================================================================

unsigned short vid_gamma_red[256];
unsigned short vid_gamma_green[256];
unsigned short vid_gamma_blue[256];

unsigned short vid_sysgamma_red[256];
unsigned short vid_sysgamma_green[256];
unsigned short vid_sysgamma_blue[256];

int vid_gammaworks;

/*
================
VID_Gamma_SetGamma -- apply gamma correction
================
*/
void VID_Gamma_SetGamma (void)
{
	if (draw_context && vid_gammaworks)
		if (SDL_SetGammaRamp (&vid_gamma_red[0], &vid_gamma_green[0], &vid_gamma_blue[0]) == -1)
			Con_Printf ("VID_Gamma_SetGamma: failed on SDL_SetGammaRamp\n");
}

/*
================
VID_Gamma_Restore -- restore system gamma
================
*/
void VID_Gamma_Restore (void)
{
	if (draw_context && vid_gammaworks)
		if (SDL_SetGammaRamp (&vid_sysgamma_red[0], &vid_sysgamma_green[0], &vid_sysgamma_blue[0]) == -1)
			Con_Printf ("VID_Gamma_Restore: failed on SDL_SetGammaRamp\n");
}

/*
================
VID_Gamma_Shutdown -- called on exit
================
*/
void VID_Gamma_Shutdown (void)
{
	VID_Gamma_Restore ();
}

/*
================
VID_Gamma_f -- callback when the cvar changes
================
*/
void VID_Gamma_f (void)
{
	static float oldgamma = -666;
	static float oldcontrast = -666;
	int i;

	if (vid_gamma.value == oldgamma && vid_contrast.value == oldcontrast)
		return;

	// bound cvars to menu range
	if (vid_gamma.value < 0.5) Cvar_SetValue ("gamma", 0.5);
	if (vid_gamma.value > 1.0) Cvar_SetValue ("gamma", 1.0);
	if (vid_contrast.value < 0.5) Cvar_SetValue ("contrast", 0.5);
	if (vid_contrast.value > 1.5) Cvar_SetValue ("contrast", 1.5);

	oldgamma = vid_gamma.value;
	oldcontrast = vid_contrast.value;

	for (i = 0; i < 256; i++)
	{
		float inf = (255 * pow ((i + 0.5) / 255.5, vid_gamma.value) + 0.5);

		inf -= 128;
		inf *= vid_contrast.value;
		inf += 128;

		vid_gamma_red[i] = CLAMP (0, (int) inf, 255) << 8;

		vid_gamma_green[i] = vid_gamma_red[i];
		vid_gamma_blue[i] = vid_gamma_red[i];
	}

	VID_Gamma_SetGamma ();
}

/*
================
VID_Gamma_Init -- call on init
================
*/
void VID_Gamma_Init (void)
{
	vid_gammaworks = false;

	if (SDL_GetGammaRamp (&vid_sysgamma_red[0], &vid_sysgamma_green[0], &vid_sysgamma_blue[0]) != -1)
		vid_gammaworks = true;

	Cvar_RegisterVariable (&vid_gamma, VID_Gamma_f); 	// same callback as below
	Cvar_RegisterVariable (&vid_contrast, VID_Gamma_f);	// same callback as above
}

/*
================
VID_SetMode
================
*/
void GL_DeleteAllLightmaps (void);
void GL_CheckExtensions (void);
void TM_DeleteTextureObjects (void);
void R_CreateParticleShaders (void);
void RBrush_LoadVertexBuffers (void);
void GLMesh_LoadVertexBuffers (void);
void GLIQM_LoadVertexBuffers (void);
void GL_BuildLightmaps (void);

int VID_SetMode (int modenum)
{
	int		temp;
	qboolean	stat = false;
	Uint32	flags = SDL_DEFAULT_FLAGS; // http://sdl.beuc.net/sdl.wiki/SDL_SetVideoMode
	char		caption[64];

	// TODO: check if video mode is supported using SDL_VideoModeOk
	if ((windowed && (modenum != 0)) ||
			(!windowed && (modenum < 1)) ||
			(!windowed && (modenum >= nummodes)))
	{
		Sys_Error ("Bad video mode\n");
	}

	// so Con_Printfs don't mess us up by forcing vid and snd updates
	temp = scr_disabled_for_loading;
	scr_disabled_for_loading = true;
	block_drawing = true;

	TM_DeleteTextureObjects ();
	CDAudio_Pause ();

	// set vertical sync (mh - made this more portable)
	if (gl_swap_control)
	{
		if (vid_vsync.value)
		{
			if (SDL_GL_SetAttribute (SDL_GL_SWAP_CONTROL, 1) == -1)
				Con_Printf ("VID_Vsync_f: failed on SDL_GL_SetAttribute\n");
		}
		else
		{
			if (SDL_GL_SetAttribute (SDL_GL_SWAP_CONTROL, 0) == -1)
				Con_Printf ("VID_Vsync_f: failed on SDL_GL_SetAttribute\n");
		}
	}

	if (modelist[modenum].type == MODE_WINDOWED)
	{
		if (_windowed_mouse.value && key_dest == key_game)
		{
			draw_context = SDL_SetVideoMode (modelist[modenum].width,
											 modelist[modenum].height,
											 modelist[modenum].bpp, flags);
			stat = true;
		}
		else
		{
			draw_context = SDL_SetVideoMode (modelist[modenum].width,
											 modelist[modenum].height,
											 modelist[modenum].bpp, flags);
			stat = true;
		}

		modestate = MODE_WINDOWED;
		// TODO set icon and title
	}
	else if (modelist[modenum].type == MODE_FULLSCREEN_DEFAULT)
	{
		flags |= SDL_FULLSCREEN;
		draw_context = SDL_SetVideoMode (modelist[modenum].width,
										 modelist[modenum].height,
										 modelist[modenum].bpp, flags);
		stat = true;
		modestate = MODE_FULLSCREEN_DEFAULT;
	}
	else
	{
		Sys_Error ("VID_SetMode: Bad mode type in modelist");
	}

	if (!stat)
	{
		Sys_Error ("Couldn't set video mode");
	}

	// kristian -- set window caption
	sprintf (caption, "RMQ Engine %1.2f.%d (%s)", (float) FITZQUAKE_VERSION, QUAKESPASM_VER_PATCH, RMQ_ENGINE_VERSION);
	SDL_WM_SetCaption ((const char *) &caption, (const char *) &caption);

	vid.width = modelist[modenum].width;
	vid.height = modelist[modenum].height;
	vid.conwidth = vid.width & 0xFFFFFFF8;
	vid.conheight = vid.conwidth * vid.height / vid.width;
	vid.numpages = 2;
	vid.type = modelist[modenum].type;

	VID_UpdateWindowStatus ();

	CDAudio_Resume ();

	vid_modenum = modenum;

	// fix the leftover Alt from any Alt-Tab or the like that switched us away
	ClearAllStates ();

	if (!msg_suppress_1)
		Con_SafePrintf ("Video mode %s initialized\n", VID_GetModeDescription (vid_modenum));

	vid.recalc_refdef = 1;

	// with SDL, this needs to be done every time the render context is recreated, so I moved it here
	GL_Init ();	// re-init because here we have a brand new render context in a brand new window so we need to assume that *EVERYTHING" is invalid...
	GL_SetupState ();

	// this deletes texture objects and PBOs used for lightmaps so that the renderer can create them on demand the next time they are needed
	// needs to be done before image reloading otherwise things go wacko (why?)
	GL_DeleteAllLightmaps ();

	// reload textures from their source files
	TM_ReloadImages ();

	// no pending changes
	vid_changed = false;

	// unbind all vertex and index arrays
	GL_UncacheArrays ();

	// unbind all textures
	GL_UnbindTextures ();

	block_drawing = false;
	scr_disabled_for_loading = temp;

	// get a batch of vertex buffers
	// we do it this way because our vbos go NUTS on a vid mode switch from windowed to fullscreen (or vice-versa) but oddly enough
	// not on a switch between 2 different modes of the same type.  SDL context handling bug???
	GL_CreateBuffers ();

	// begin vertex buffers
	GL_BeginBuffers ();

	RBrush_LoadVertexBuffers ();
	GLMesh_LoadVertexBuffers ();
	GLIQM_LoadVertexBuffers ();

	// rebuild lightmaps
	GL_BuildLightmaps ();

	R_CreateRTTTextures (modelist[modenum].width, modelist[modenum].height);

	return true;
}

/*
===================
VID_Changed_f -- kristian -- notify us that a value has changed that requires a vid_restart
===================
*/
void VID_Changed_f (void)
{
	vid_changed = true;
}

/*
===================
VID_Restart -- johnfitz -- change video modes on the fly
===================
*/
void VID_Restart (void)
{
	int		i;
	vmode_t		oldmode;

	if (vid_locked || !vid_changed)
		return;

	// decide which mode to set
	oldmode = modelist[vid_default];

	if (vid_fullscreen.value)
	{
		for (i = 1; i < nummodes; i++)
		{
			if (modelist[i].width == (int) vid_width.value &&
				modelist[i].height == (int) vid_height.value &&
				modelist[i].bpp == (int) vid_bpp.value)
			{
				break;
			}
		}

		if (i == nummodes)
		{
			Con_Printf ("%dx%dx%d %dHz is not a valid fullscreen mode\n",
						(int) vid_width.value,
						(int) vid_height.value,
						(int) vid_bpp.value,
						(int) vid_refreshrate.value);
			return;
		}

		windowed = false;
		vid_default = i;
	}
	else // not fullscreen
	{
		if (vid_width.value < 640)
		{
			Con_Printf ("Window width can't be less than 640\n");
			return;
		}

		if (vid_height.value < 480)
		{
			Con_Printf ("Window height can't be less than 480\n");
			return;
		}

		modelist[0].width = (int) vid_width.value;
		modelist[0].height = (int) vid_height.value;
		sprintf (modelist[0].modedesc, "%dx%dx%d",
				 modelist[0].width,
				 modelist[0].height,
				 modelist[0].bpp);

		windowed = true;
		vid_default = 0;
	}

	// set new mode
	VID_SetMode (vid_default);

	vid_canalttab = true;

	// conwidth and conheight need to be recalculated
	vid.conwidth = (scr_conwidth.value > 0) ? (int) scr_conwidth.value : (scr_conscale.value > 0) ? (int) (vid.width / scr_conscale.value) : vid.width;
	vid.conwidth = CLAMP (320, vid.conwidth, vid.width);
	vid.conwidth &= 0xFFFFFFF8;
	vid.conheight = vid.conwidth * vid.height / vid.width;

	// keep cvars in line with actual mode
	Cvar_Set ("vid_width", va ("%i", modelist[vid_default].width));
	Cvar_Set ("vid_height", va ("%i", modelist[vid_default].height));
	Cvar_Set ("vid_bpp", va ("%i", modelist[vid_default].bpp));
	Cvar_Set ("vid_fullscreen", (windowed) ? "0" : "1");
}


void VID_Restart_f (void)
{
	// force a vid-restart from the command-line
	// vid_changed = true;
	VID_Restart ();
}


void VID_ForceRestart_f (void)
{
	// force a vid-restart from the command-line
	vid_changed = true;
	VID_Restart ();
}


/*
================
VID_Test -- johnfitz -- like vid_restart, but asks for confirmation after switching modes
================
*/
void VID_Test (void)
{
	vmode_t oldmode;

	if (vid_locked || !vid_changed)
		return;

	// now try the switch
	oldmode = modelist[vid_default];

	VID_Restart ();

	// pop up confirmation dialoge
	if (!SCR_ModalMessage ("Would you like to keep this\nvideo mode? (y/n)\n", 5.0f))
	{
		// revert cvars and mode
		Cvar_Set ("vid_width", va ("%i", oldmode.width));
		Cvar_Set ("vid_height", va ("%i", oldmode.height));
		Cvar_Set ("vid_bpp", va ("%i", oldmode.bpp));
		Cvar_Set ("vid_fullscreen", (oldmode.type == MODE_WINDOWED) ? "0" : "1");
		VID_Restart ();
	}
}

/*
================
VID_Unlock -- johnfitz
================
*/
void VID_Unlock (void)
{
	vid_locked = false;

	// sync up cvars in case they were changed during the lock
	Cvar_Set ("vid_width", va ("%i", modelist[vid_default].width));
	Cvar_Set ("vid_height", va ("%i", modelist[vid_default].height));
	Cvar_Set ("vid_bpp", va ("%i", modelist[vid_default].bpp));
	Cvar_Set ("vid_fullscreen", (windowed) ? "0" : "1");
}

/*
================
VID_UpdateWindowStatus
================
*/
void VID_UpdateWindowStatus (void)
{
	// IN_UpdateClipCursor ();
}

// ==============================================================================
//
// 	OPENGL STUFF
//
// ==============================================================================

/*
===============
GL_Info_f -- johnfitz
===============
*/
void GL_Info_f (void)
{
	Con_SafePrintf ("GL_VENDOR: %s\n", gl_vendor);
	Con_SafePrintf ("GL_RENDERER: %s\n", gl_renderer);
	Con_SafePrintf ("GL_VERSION: %s\n", gl_version);
	VID_PrintGLExtensions ();
}


void Draw_CreateShaders (void);
void Warp_CreateShaders (void);
void Sky_CreateShaders (void);
void Part_CreateShaders (void);
void Mesh_CreateShaders (void);
void IQM_CreateShaders (void);
void Main_CreateShaders (void);
void World_CreateShaders (void);
void Sprite_CreateShaders (void);
void Corona_CreateShaders (void);


qboolean GL_GetGLSLExtensions (void)
{
	// let's use the proper OpenGL 2.0 extensions here
	if (!Q_strstr (gl_extensions, "GL_ARB_vertex_program ")) return false;
	if (!Q_strstr (gl_extensions, "GL_ARB_fragment_program ")) return false;

	if (!(qglBindProgramARB = (PFNGLBINDPROGRAMARBPROC) SDL_GL_GetProcAddress ("glBindProgramARB"))) return false;
	if (!(qglDeleteProgramsARB = (PFNGLDELETEPROGRAMSARBPROC) SDL_GL_GetProcAddress ("glDeleteProgramsARB"))) return false;
	if (!(qglGenProgramsARB = (PFNGLGENPROGRAMSARBPROC) SDL_GL_GetProcAddress ("glGenProgramsARB"))) return false;
	if (!(qglGetProgramEnvParameterdvARB = (PFNGLGETPROGRAMENVPARAMETERDVARBPROC) SDL_GL_GetProcAddress ("glGetProgramEnvParameterdvARB"))) return false;
	if (!(qglGetProgramEnvParameterfvARB = (PFNGLGETPROGRAMENVPARAMETERFVARBPROC) SDL_GL_GetProcAddress ("glGetProgramEnvParameterfvARB"))) return false;
	if (!(qglGetProgramLocalParameterdvARB = (PFNGLGETPROGRAMLOCALPARAMETERDVARBPROC) SDL_GL_GetProcAddress ("glGetProgramLocalParameterdvARB"))) return false;
	if (!(qglGetProgramLocalParameterfvARB = (PFNGLGETPROGRAMLOCALPARAMETERFVARBPROC) SDL_GL_GetProcAddress ("glGetProgramLocalParameterfvARB"))) return false;
	if (!(qglGetProgramStringARB = (PFNGLGETPROGRAMSTRINGARBPROC) SDL_GL_GetProcAddress ("glGetProgramStringARB"))) return false;
	if (!(qglGetProgramivARB = (PFNGLGETPROGRAMIVARBPROC) SDL_GL_GetProcAddress ("glGetProgramivARB"))) return false;
	if (!(qglIsProgramARB = (PFNGLISPROGRAMARBPROC) SDL_GL_GetProcAddress ("glIsProgramARB"))) return false;
	if (!(qglProgramEnvParameter4dARB = (PFNGLPROGRAMENVPARAMETER4DARBPROC) SDL_GL_GetProcAddress ("glProgramEnvParameter4dARB"))) return false;
	if (!(qglProgramEnvParameter4dvARB = (PFNGLPROGRAMENVPARAMETER4DVARBPROC) SDL_GL_GetProcAddress ("glProgramEnvParameter4dvARB"))) return false;
	if (!(qglProgramEnvParameter4fARB = (PFNGLPROGRAMENVPARAMETER4FARBPROC) SDL_GL_GetProcAddress ("glProgramEnvParameter4fARB"))) return false;
	if (!(qglProgramEnvParameter4fvARB = (PFNGLPROGRAMENVPARAMETER4FVARBPROC) SDL_GL_GetProcAddress ("glProgramEnvParameter4fvARB"))) return false;
	if (!(qglProgramLocalParameter4dARB = (PFNGLPROGRAMLOCALPARAMETER4DARBPROC) SDL_GL_GetProcAddress ("glProgramLocalParameter4dARB"))) return false;
	if (!(qglProgramLocalParameter4dvARB = (PFNGLPROGRAMLOCALPARAMETER4DVARBPROC) SDL_GL_GetProcAddress ("glProgramLocalParameter4dvARB"))) return false;
	if (!(qglProgramLocalParameter4fARB = (PFNGLPROGRAMLOCALPARAMETER4FARBPROC) SDL_GL_GetProcAddress ("glProgramLocalParameter4fARB"))) return false;
	if (!(qglProgramLocalParameter4fvARB = (PFNGLPROGRAMLOCALPARAMETER4FVARBPROC) SDL_GL_GetProcAddress ("glProgramLocalParameter4fvARB"))) return false;
	if (!(qglProgramStringARB = (PFNGLPROGRAMSTRINGARBPROC) SDL_GL_GetProcAddress ("glProgramStringARB"))) return false;
	if (!(qglVertexAttribPointerARB = (PFNGLVERTEXATTRIBPOINTERARBPROC) SDL_GL_GetProcAddress ("glVertexAttribPointerARB"))) return false;
	if (!(qglEnableVertexAttribArrayARB = (PFNGLENABLEVERTEXATTRIBARRAYARBPROC) SDL_GL_GetProcAddress ("glEnableVertexAttribArrayARB"))) return false;
	if (!(qglDisableVertexAttribArrayARB = (PFNGLDISABLEVERTEXATTRIBARRAYARBPROC) SDL_GL_GetProcAddress ("glDisableVertexAttribArrayARB"))) return false;

	// these are optional so we don't want to print a warning if they don't exist
	Con_Printf ("FOUND: GL_ARB_vertex_program\n");
	Con_Printf ("FOUND: GL_ARB_fragment_program\n");

	// clear down anything we had
	R_DeleteShaders ();

	// now load them up again
	Warp_CreateShaders ();
	Draw_CreateShaders ();
	Sky_CreateShaders ();
	Part_CreateShaders ();
	Mesh_CreateShaders ();
	IQM_CreateShaders ();
	Main_CreateShaders ();
	World_CreateShaders ();
	Sprite_CreateShaders ();
	Corona_CreateShaders ();

	// now enable our shaders
	glEnable (GL_VERTEX_PROGRAM_ARB);
	glEnable (GL_FRAGMENT_PROGRAM_ARB);

	// bind the default shaders (also updates cached state)
	GL_SetPrograms (0, 0);
	return true;
}


qboolean gl_primitive_restart = false;
qboolean gl_prenabled = false;

// both the core and the NV versions have the same signature
PFNGLPRIMITIVERESTARTINDEXPROC qglPrimitiveRestartIndex = NULL;
PFNGLPRIMITIVERESTARTINDEXPROC qglPrimitiveRestartIndexNV = NULL;

// grrrr - some total weenie must have gotten immense satisfaction out of having created such a
// technically perfect but awkward to use contraption
void GL_EnablePrimitiveRestart (qboolean enable)
{
	if (!gl_primitive_restart) return;

	if (gl_prenabled && !enable)
	{
		if (qglPrimitiveRestartIndex)
			glDisable (GL_PRIMITIVE_RESTART);
		else glDisable (GL_PRIMITIVE_RESTART_NV);

		gl_prenabled = false;
	}
	else if (!gl_prenabled && enable)
	{
		if (qglPrimitiveRestartIndex)
		{
			glEnable (GL_PRIMITIVE_RESTART);
			qglPrimitiveRestartIndex (0xffff);
		}
		else
		{
			glEnable (GL_PRIMITIVE_RESTART_NV);
			qglPrimitiveRestartIndexNV (0xffff);
		}

		gl_prenabled = true;
	}
}

// CVA replacment funcs so that I can keep a consistent API
void APIENTRY glLockArraysRMQ (GLint blah1, GLsizei blah2) {}
void APIENTRY glUnlockArraysRMQ (void) {}

// replacement for glDrawRangeElements so that I can keep a consistent API
void APIENTRY glDrawRangeElementsRMQ (GLenum mode, GLuint start, GLuint end, GLsizei count, GLenum type, const GLvoid *indices)
{
	glDrawElements (mode, count, type, indices);
}


void GL_GetCVAExtensions (void)
{
	if (Q_strstr (gl_extensions, "GL_EXT_compiled_vertex_array "))
	{
		qglLockArrays = (PFNGLLOCKARRAYSEXTPROC) SDL_GL_GetProcAddress ("glLockArraysEXT");
		qglUnlockArrays = (PFNGLUNLOCKARRAYSEXTPROC) SDL_GL_GetProcAddress ("glUnlockArraysEXT");

		if (qglLockArrays && qglUnlockArrays)
		{
			gl_ext_compiled_vertex_array = true;
			Con_Printf ("FOUND: EXT_compiled_vertex_array\n");
		}
		else Con_Warning ("GL_EXT_compiled_vertex_array not supported (GetProcAddress failed)\n");
	}
	else Con_Warning ("GL_EXT_compiled_vertex_array not supported\n");

	// set up our replacement CVA functions
	if (!qglLockArrays || !qglUnlockArrays)
	{
		qglLockArrays = glLockArraysRMQ;
		qglUnlockArrays = glUnlockArraysRMQ;
	}
}


void GL_GetBufferExtensions (void)
{
	qglBindBuffer = (PFNGLBINDBUFFERPROC) SDL_GL_GetProcAddress ("glBindBufferARB");
	qglDeleteBuffers = (PFNGLDELETEBUFFERSPROC) SDL_GL_GetProcAddress ("glDeleteBuffersARB");
	qglGenBuffers = (PFNGLGENBUFFERSPROC) SDL_GL_GetProcAddress ("glGenBuffersARB");
	qglIsBuffer = (PFNGLISBUFFERPROC) SDL_GL_GetProcAddress ("glIsBufferARB");
	qglBufferData = (PFNGLBUFFERDATAPROC) SDL_GL_GetProcAddress ("glBufferDataARB");
	qglBufferSubData = (PFNGLBUFFERSUBDATAPROC) SDL_GL_GetProcAddress ("glBufferSubDataARB");
	qglGetBufferSubData = (PFNGLGETBUFFERSUBDATAPROC) SDL_GL_GetProcAddress ("glGetBufferSubDataARB");
	qglMapBuffer = (PFNGLMAPBUFFERPROC) SDL_GL_GetProcAddress ("glMapBufferARB");
	qglUnmapBuffer = (PFNGLUNMAPBUFFERPROC) SDL_GL_GetProcAddress ("glUnmapBufferARB");
	qglGetBufferParameteriv = (PFNGLGETBUFFERPARAMETERIVPROC) SDL_GL_GetProcAddress ("glGetBufferParameterivARB");
	qglGetBufferPointerv = (PFNGLGETBUFFERPOINTERVPROC) SDL_GL_GetProcAddress ("glGetBufferPointervARB");

	if (qglBindBuffer && qglDeleteBuffers && qglGenBuffers && qglIsBuffer &&
		qglBufferData && qglBufferSubData && qglGetBufferSubData && qglMapBuffer &&
		qglUnmapBuffer && qglGetBufferParameteriv && qglGetBufferPointerv)
	{
		Con_Printf ("FOUND: ARB_vertex_buffer_object\n");
	}
	else
	{
		Sys_Error ("GL_ARB_vertex_buffer_object not available");
	}
}


void GL_GetQueryExtensions (void)
{
	// occlusion queries
	qglGenQueries = (PFNGLGENQUERIESPROC) SDL_GL_GetProcAddress ("glGenQueriesARB");
	qglDeleteQueries = (PFNGLDELETEQUERIESPROC) SDL_GL_GetProcAddress ("glDeleteQueriesARB");
	qglIsQuery = (PFNGLISQUERYPROC) SDL_GL_GetProcAddress ("glIsQueryARB");
	qglBeginQuery = (PFNGLBEGINQUERYPROC) SDL_GL_GetProcAddress ("glBeginQueryARB");
	qglEndQuery = (PFNGLENDQUERYPROC) SDL_GL_GetProcAddress ("glEndQueryARB");
	qglGetQueryiv = (PFNGLGETQUERYIVPROC) SDL_GL_GetProcAddress ("glGetQueryivARB");
	qglGetQueryObjectiv = (PFNGLGETQUERYOBJECTIVPROC) SDL_GL_GetProcAddress ("glGetQueryObjectivARB");
	qglGetQueryObjectuiv = (PFNGLGETQUERYOBJECTUIVPROC) SDL_GL_GetProcAddress ("glGetQueryObjectuivARB");
}


void GL_GetMiscExtensions (void)
{
	// set up our replacement DrawRangeElements function
	if (!(qglDrawRangeElements = (PFNGLDRAWRANGEELEMENTSPROC) SDL_GL_GetProcAddress ("glDrawRangeElements")))
		qglDrawRangeElements = glDrawRangeElementsRMQ;

	qglMultiDrawArrays = (PFNGLMULTIDRAWARRAYSPROC) SDL_GL_GetProcAddress ("glMultiDrawArrays");
	qglBlendEquation = (PFNGLBLENDEQUATIONPROC) SDL_GL_GetProcAddress ("glBlendEquation");
	qglSecondaryColorPointer = (PFNGLSECONDARYCOLORPOINTERPROC) SDL_GL_GetProcAddress ("glSecondaryColorPointer");

	qglPrimitiveRestartIndex = (PFNGLPRIMITIVERESTARTINDEXPROC) SDL_GL_GetProcAddress ("glPrimitiveRestartIndex");
	qglPrimitiveRestartIndexNV = (PFNGLPRIMITIVERESTARTINDEXPROC) SDL_GL_GetProcAddress ("glPrimitiveRestartIndexNV");

	if (qglPrimitiveRestartIndex || qglPrimitiveRestartIndexNV)
	{
		gl_primitive_restart = true;
		gl_prenabled = false;
		Con_Printf ("FOUND: primitive restart\n");
	}
}


void GL_GetTextureExtensions (void)
{
	// mh - if the GL_VERSION is high enough we don't need to bother looking for these; we can just assume that they're present
	// multitexture
	if (Q_strstr (gl_extensions, "GL_ARB_multitexture "))
	{
		qglMultiTexCoord2fARB = (PFNGLMULTITEXCOORD2FARBPROC) SDL_GL_GetProcAddress ("glMultiTexCoord2fARB");
		qglActiveTexture = (PFNGLACTIVETEXTUREARBPROC) SDL_GL_GetProcAddress ("glActiveTextureARB");
		qglClientActiveTexture = (PFNGLCLIENTACTIVETEXTUREARBPROC) SDL_GL_GetProcAddress ("glClientActiveTextureARB");

		if (qglMultiTexCoord2fARB && qglActiveTexture && qglClientActiveTexture)
			Con_Printf ("FOUND: ARB_multitexture\n");
		else Sys_Error ("multitexture not supported (GetProcAddress failed)\n");
	}
	else Sys_Error ("multitexture not supported (extension not found)\n");

	if (Q_strstr (gl_extensions, "GL_ARB_texture_compression "))
	{
		Con_Printf ("FOUND: ARB_texture_compression\n");
		gl_arb_texture_compression = true;
	}
	else
	{
		Con_Warning ("ARB_texture_compression not supported\n");
		gl_arb_texture_compression = false;
	}

	if (Q_strstr (gl_extensions, "GL_ARB_texture_cube_map "))
		Con_Printf ("FOUND: ARB_texture_cube_map\n");
	else Sys_Error ("ARB_texture_cube_map not supported\n");

	// texture_env_combine
	if (Q_strstr (gl_extensions, "GL_ARB_texture_env_combine "))
		Con_Printf ("FOUND: ARB_texture_env_combine\n");
	else if (Q_strstr (gl_extensions, "GL_EXT_texture_env_combine "))
		Con_Printf ("FOUND: EXT_texture_env_combine\n");
	else Sys_Error ("texture_env_combine not supported\n");

	// texture_env_add
	if (Q_strstr (gl_extensions, "GL_ARB_texture_env_add "))
		Con_Printf ("FOUND: ARB_texture_env_add\n");
	else if (Q_strstr (gl_extensions, "GL_EXT_texture_env_add "))
		Con_Printf ("FOUND: EXT_texture_env_add\n");
	else Sys_Error ("texture_env_add not supported\n");

	// anisotropic filtering
	if (Q_strstr (gl_extensions, "GL_EXT_texture_filter_anisotropic "))
	{
		float test2;
		GLuint tex;

		// test to make sure we really have control over it
		// 1.0 and 2.0 should always be legal values
		// (1 is not - the minimum to support the spec is 2)
		// http://www.opengl.org/registry/specs/EXT/texture_filter_anisotropic.txt
		glGenTextures (1, &tex);
		glBindTexture (GL_TEXTURE_2D, tex);
		glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, 2.0f);
		glGetTexParameterfv (GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, &test2);
		glDeleteTextures (1, &tex);

		if (test2 == 2)
		{
			Con_Printf ("FOUND: EXT_texture_filter_anisotropic\n");
			gl_ext_texture_filter_anisotropic = true;
		}
		else Con_Warning ("anisotropic filtering locked by driver. Current driver setting is %f\n", test2);

		// get max value either way, so the menu and stuff know it
		glGetFloatv (GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &gl_max_anisotropy);
	}
	else Con_Warning ("texture_filter_anisotropic not supported\n");
}


void GL_GetVSyncExtensions (void)
{
	int swap_control;

	// swap control
	if (Q_strstr (gl_extensions, "GL_EXT_swap_control "))
	{
		if (SDL_GL_SetAttribute (SDL_GL_SWAP_CONTROL, 0) == -1)
		{
			Con_Warning ("vertical sync not supported (SDL_GL_SetAttribute failed)\n");
		}
		else
		{
			if (SDL_GL_GetAttribute (SDL_GL_SWAP_CONTROL, &swap_control) == -1)
			{
				Con_Warning ("vertical sync not supported (SDL_GL_GetAttribute failed).\nMake sure you don't have vertical sync disabled in your driver settings.\n");
			}
			else if (swap_control == -1)
			{
				// TODO: check if this is correct - I don't know what SDL returns if vertical sync is disabled
				Con_Warning ("vertical sync not supported (swap interval is -1.)\nMake sure you don't have vertical sync disabled in your driver settings.\n");
			}
			else
			{
				Con_Printf ("FOUND: GL_EXT_swap_control\n");
				gl_swap_control = true;
			}
		}
	}
	else Con_Warning ("vertical sync not supported (extension not found)\n");
}


/*
===============
GL_CheckExtensions -- johnfitz

mh - multiple changes here

why the fuck don't SGI/ARB/Khronos/whoever just release proper headers and libs with a well-defined API for
easily checking availability of a cap?  This shit is awful.

SIGH - some hardware and OS combinations kick and scream unless you specifically get the -ARB or whatever shit versions
of entrypoints.  OpenGL drivers still suck, my friend.
===============
*/
void GL_CheckExtensions (void)
{
	if (!gl_extensions)
	{
		Sys_Error ("glGetString (GL_EXTENSIONS) returned NULL!");
		return;
	}

	GL_GetMiscExtensions ();
	GL_GetTextureExtensions ();
	GL_GetCVAExtensions ();
	GL_GetQueryExtensions ();
	GL_GetBufferExtensions ();
	GL_GetVSyncExtensions ();

	if (!GL_GetGLSLExtensions ())
		Sys_Error ("Could not load vertex/fragment program extensions!");
}


/*
===============
GL_SetupState -- johnfitz

does all the stuff from GL_Init that needs to be done every time a new GL render context is created
GL_Init will still do the stuff that only needs to be done once
===============
*/

int Q_SDL_GL_RED_SIZE = 0;
int Q_SDL_GL_GREEN_SIZE = 0;
int Q_SDL_GL_BLUE_SIZE = 0;
int Q_SDL_GL_ALPHA_SIZE = 0;
int Q_SDL_GL_BUFFER_SIZE = 0;
int Q_SDL_GL_DOUBLEBUFFER = 0;
int Q_SDL_GL_DEPTH_SIZE = 0;
int Q_SDL_GL_STENCIL_SIZE = 0;
int Q_SDL_GL_ACCUM_RED_SIZE = 0;
int Q_SDL_GL_ACCUM_GREEN_SIZE = 0;
int Q_SDL_GL_ACCUM_BLUE_SIZE = 0;
int Q_SDL_GL_ACCUM_ALPHA_SIZE = 0;
int Q_SDL_GL_STEREO = 0;
int Q_SDL_GL_MULTISAMPLEBUFFERS = 0;
int Q_SDL_GL_MULTISAMPLESAMPLES = 0;
int Q_SDL_GL_ACCELERATED_VISUAL = 0;
int Q_SDL_GL_SWAP_CONTROL = 0;

void GL_SetupState (void)
{
	glClearColor (0.15, 0.15, 0.15, 0); // johnfitz -- originally 1,0,0,0

	glCullFace (GL_FRONT); // GL_BACK); // johnfitz -- glquake used CCW with backwards culling -- let's do it right
	// glFrontFace (GL_CW); // johnfitz -- glquake used CCW with backwards culling -- let's do it right

	glEnable (GL_ALPHA_TEST);
	glAlphaFunc (GL_GREATER, 0.666);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDepthFunc (GL_LEQUAL);

	glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
	glShadeModel (GL_FLAT);

	glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST); // johnfitz

	glDisable (GL_LIGHTING);
	glDisable (GL_DITHER);

	// read back all of our attributes
	SDL_GL_GetAttribute (SDL_GL_RED_SIZE, &Q_SDL_GL_RED_SIZE);
	SDL_GL_GetAttribute (SDL_GL_GREEN_SIZE, &Q_SDL_GL_GREEN_SIZE);
	SDL_GL_GetAttribute (SDL_GL_BLUE_SIZE, &Q_SDL_GL_BLUE_SIZE);
	SDL_GL_GetAttribute (SDL_GL_ALPHA_SIZE, &Q_SDL_GL_ALPHA_SIZE);
	SDL_GL_GetAttribute (SDL_GL_BUFFER_SIZE, &Q_SDL_GL_BUFFER_SIZE);
	SDL_GL_GetAttribute (SDL_GL_DOUBLEBUFFER, &Q_SDL_GL_DOUBLEBUFFER);
	SDL_GL_GetAttribute (SDL_GL_DEPTH_SIZE, &Q_SDL_GL_DEPTH_SIZE);
	SDL_GL_GetAttribute (SDL_GL_STENCIL_SIZE, &Q_SDL_GL_STENCIL_SIZE);
	SDL_GL_GetAttribute (SDL_GL_ACCUM_RED_SIZE, &Q_SDL_GL_ACCUM_RED_SIZE);
	SDL_GL_GetAttribute (SDL_GL_ACCUM_GREEN_SIZE, &Q_SDL_GL_ACCUM_GREEN_SIZE);
	SDL_GL_GetAttribute (SDL_GL_ACCUM_BLUE_SIZE, &Q_SDL_GL_ACCUM_BLUE_SIZE);
	SDL_GL_GetAttribute (SDL_GL_ACCUM_ALPHA_SIZE, &Q_SDL_GL_ACCUM_ALPHA_SIZE);
	SDL_GL_GetAttribute (SDL_GL_STEREO, &Q_SDL_GL_STEREO);
	SDL_GL_GetAttribute (SDL_GL_MULTISAMPLEBUFFERS, &Q_SDL_GL_MULTISAMPLEBUFFERS);
	SDL_GL_GetAttribute (SDL_GL_MULTISAMPLESAMPLES, &Q_SDL_GL_MULTISAMPLESAMPLES);
	SDL_GL_GetAttribute (SDL_GL_ACCELERATED_VISUAL, &Q_SDL_GL_ACCELERATED_VISUAL);
	SDL_GL_GetAttribute (SDL_GL_SWAP_CONTROL, &Q_SDL_GL_SWAP_CONTROL);
}


/*
===============
GL_Init
===============
*/
void GL_Init (void)
{
	// clear to black
	glClearColor (0, 0, 0, 1);
	glClear (GL_COLOR_BUFFER_BIT);
	SDL_GL_SwapBuffers ();

	gl_vendor = (const char *) glGetString (GL_VENDOR);
	gl_renderer = (const char *) glGetString (GL_RENDERER);
	gl_version = (const char *) glGetString (GL_VERSION);
	gl_extensions = (const char *) glGetString (GL_EXTENSIONS);

	// print these always as they're valuable info for debugging
	Con_SafePrintf ("GL_VENDOR: %s\n", gl_vendor);
	Con_SafePrintf ("GL_RENDERER: %s\n", gl_renderer);
	Con_SafePrintf ("GL_VERSION: %s\n", gl_version);
	// VID_PrintGLExtensions ();

	// assume we have no vendor specific behavior needed (...yet...!)
	gl_brokendrivers.AMDATIBroken = false;
	gl_brokendrivers.NVIDIABroken = false;
	gl_brokendrivers.IntelBroken = false;

	// to do - AMD/ATI string???
	if (Q_strstr (gl_vendor, "Intel")) gl_brokendrivers.IntelBroken = true;
	if (Q_strstr (gl_vendor, "NVIDIA")) gl_brokendrivers.NVIDIABroken = true;

	GL_CheckExtensions (); // johnfitz

#if 1
	// johnfitz -- confirm presence of stencil buffer
	glGetIntegerv (GL_STENCIL_BITS, &gl_stencilbits);

	if (!gl_stencilbits)
		Con_Warning ("Could not create stencil buffer\n");
	else
	{
		Con_Printf ("%i bit stencil buffer\n", gl_stencilbits);
		glClearStencil (1);
	}
#endif

	GL_SetupState (); // johnfitz
}

/*
=================
GL_BeginRendering -- sets values of glwidth, glheight

mh - glx and gly were always 0 and 0 so they were gotten rid of
this really should pass a struct...
=================
*/
void GL_BeginRendering (viddef_t *vdef)
{
	vdef->glwidth = vid.width;
	vdef->glheight = vid.height;
}


/*
=================
GL_EndRendering
=================
*/
void GL_EndRendering (void)
{
	extern int scr_framecount;
	extern cvar_t gl_finish;

	if (!scr_skipupdate || block_drawing)
	{
		if (gl_finish.value) glFinish ();

		SDL_GL_SwapBuffers ();
		scr_framecount++;
	}
}


void VID_SetDefaultMode (void)
{
}


void VID_Shutdown (void)
{
	if (vid_initialized)
	{
		R_DeleteShaders ();
		vid_canalttab = false;
		VID_Gamma_Shutdown (); // johnfitz

		SDL_QuitSubSystem (SDL_INIT_VIDEO);
		draw_context = NULL;

		PL_VID_Shutdown ();
	}
}

// ==========================================================================

/*
===================================================================

MAIN WINDOW

===================================================================
*/

/*
================
ClearAllStates
================
*/
void ClearAllStates (void)
{
	int		i;

	// send an up event for each key, to make sure the server clears them all
	for (i = 0; i < 256; i++)
	{
		Key_Event (i, false);
	}

	Key_ClearStates ();
	IN_ClearStates ();
}



// ==========================================================================
//
//  COMMANDS
//
// ==========================================================================

/*
=================
VID_NumModes
=================
*/
int VID_NumModes (void)
{
	return nummodes;
}


/*
=================
VID_GetModePtr
=================
*/
vmode_t *VID_GetModePtr (int modenum)
{

	if ((modenum >= 0) && (modenum < nummodes))
		return &modelist[modenum];
	else
		return &badmode;
}


/*
=================
VID_GetModeDescription
=================
*/
char *VID_GetModeDescription (int mode)
{
	char		*pinfo;
	vmode_t		*pv;
	static char	temp[100];

	if ((mode < 0) || (mode >= nummodes))
		return NULL;

	if (!leavecurrentmode)
	{
		pv = VID_GetModePtr (mode);
		pinfo = pv->modedesc;
	}
	else
	{
		sprintf (temp, "Desktop resolution (%ix%ix%i)", // johnfitz -- added bpp
				 modelist[MODE_FULLSCREEN_DEFAULT].width,
				 modelist[MODE_FULLSCREEN_DEFAULT].height,
				 modelist[MODE_FULLSCREEN_DEFAULT].bpp); // johnfitz -- added bpp
		pinfo = temp;
	}

	return pinfo;
}

// KJB: Added this to return the mode driver name in description for console
/*
=================
VID_GetExtModeDescription
=================
*/
char *VID_GetExtModeDescription (int mode)
{
	static char	pinfo[40];
	vmode_t		*pv;

	if ((mode < 0) || (mode >= nummodes))
		return NULL;

	pv = VID_GetModePtr (mode);

	if (modelist[mode].type == MODE_FULLSCREEN_DEFAULT)
	{
		if (!leavecurrentmode)
		{
			sprintf (pinfo, "%s fullscreen", pv->modedesc);
		}
		else
		{
			sprintf (pinfo, "Desktop resolution (%ix%ix%i)", // johnfitz -- added bpp
					 modelist[MODE_FULLSCREEN_DEFAULT].width,
					 modelist[MODE_FULLSCREEN_DEFAULT].height,
					 modelist[MODE_FULLSCREEN_DEFAULT].bpp); // johnfitz -- added bpp
		}
	}
	else
	{
		if (modestate == MODE_WINDOWED)
			sprintf (pinfo, "%s windowed", pv->modedesc);
		else
			sprintf (pinfo, "windowed");
	}

	return pinfo;
}

/*
=================
VID_DescribeCurrentMode_f
=================
*/
void VID_DescribeCurrentMode_f (void)
{
	Con_Printf ("%s\n", VID_GetExtModeDescription (vid_modenum));
}

/*
=================
VID_DescribeModes_f -- johnfitz -- changed formatting, and added refresh rates after each mode.
=================
*/
void VID_DescribeModes_f (void)
{
	int		i, lnummodes, t;
	vmode_t		*pv;
	int		lastwidth = 0, lastheight = 0, lastbpp = 0, count = 0;

	lnummodes = VID_NumModes ();

	t = leavecurrentmode;
	leavecurrentmode = 0;

	for (i = 1; i < lnummodes; i++)
	{
		pv = VID_GetModePtr (i);

		if (lastwidth != pv->width || lastheight != pv->height || lastbpp != pv->bpp)
		{
			if (count > 0)
				Con_SafePrintf ("\n");

			Con_SafePrintf ("   %4i x %4i x %i", pv->width, pv->height, pv->bpp);
			lastwidth = pv->width;
			lastheight = pv->height;
			lastbpp = pv->bpp;
			count++;
		}
	}

	Con_Printf ("\n%i modes\n", count);

	leavecurrentmode = t;
}

// ==========================================================================
//
//  INIT
//
// ==========================================================================

/*
=================
VID_InitDIB
=================
*/
void VID_InitDIB (void)
{
	const SDL_VideoInfo *info;

	modelist[0].type = MODE_WINDOWED;

	if (COM_CheckParm ("-width"))
		modelist[0].width = Q_atoi (com_argv[COM_CheckParm ("-width") +1]);
	else
		modelist[0].width = 800;	// QuakeSpasm, was 640

	if (modelist[0].width < 320)
		modelist[0].width = 320;

	if (COM_CheckParm ("-height"))
		modelist[0].height = Q_atoi (com_argv[COM_CheckParm ("-height") +1]);
	else
		modelist[0].height = modelist[0].width * 240 / 320;

	if (modelist[0].height < 200) // johnfitz -- was 240
		modelist[0].height = 200; // johnfitz -- was 240

	info = SDL_GetVideoInfo ();
	modelist[0].bpp = info->vfmt->BitsPerPixel;

	sprintf (modelist[0].modedesc, "%dx%dx%d", // johnfitz -- added bpp
			 modelist[0].width,
			 modelist[0].height,
			 modelist[0].bpp); // johnfitz -- added bpp

	modelist[0].modenum = MODE_WINDOWED;
	modelist[0].dib = 1;
	modelist[0].fullscreen = 0;
	modelist[0].halfscreen = 0;

	nummodes = 1;
}


int VID_ModeSortFunc (SDL_Rect **m1, SDL_Rect **m2)
{
	if (m1[0]->w == m2[0]->w)
		return (m1[0]->h - m2[0]->h);
	else return (m1[0]->w - m2[0]->w);
}


int VID_ModeSortFunc2 (vmode_t *m1, vmode_t *m2)
{
	if (m1->bpp == m2->bpp)
	{
		if (m1->width == m2->width)
			return (m1->height - m2->height);
		else return (m1->width - m2->width);
	}
	return (m1->bpp - m2->bpp);
}


typedef struct
{
	int			width;
	int			height;
} lmode_t;

lmode_t	lowresmodes[] =
{
	{320, 200},
	{320, 240},
	{400, 300},
	{512, 384},
	{640, 400}
};


qboolean VID_DoesModeExist (int originalnummodes, int newnummodes, int modewidth, int modeheight, int modebpp)
{
	int k;

	for (k = originalnummodes; k < newnummodes; k++)
	{
		if ((modewidth == modelist[k].width) &&
			(modeheight == modelist[k].height) &&
			(modebpp == modelist[k].bpp))
		{
			// mode already exists
			return true;
		}
	}

	// definitely a new mode
	return false;
}


void VID_AddMode (int width, int height, int bpp, qboolean warn)
{
	modelist[nummodes].type = MODE_FULLSCREEN_DEFAULT;
	modelist[nummodes].width = width;
	modelist[nummodes].height = height;
	modelist[nummodes].modenum = 0;
	modelist[nummodes].halfscreen = 0;
	modelist[nummodes].dib = 1;
	modelist[nummodes].fullscreen = 1;
	modelist[nummodes].bpp = bpp;
	modelist[nummodes].warn = warn;

	sprintf (modelist[nummodes].modedesc, "%dx%dx%d",
			 modelist[nummodes].width,
			 modelist[nummodes].height,
			 modelist[nummodes].bpp); // johnfitz -- refreshrate

	// if the width is more than twice the height, reduce it by half because this
	// is probably a dual-screen monitor
	if (!COM_CheckParm ("-noadjustaspect"))
	{
		if (modelist[nummodes].width > (modelist[nummodes].height << 1))
		{
			modelist[nummodes].width >>= 1;
			modelist[nummodes].halfscreen = 1;
			sprintf (modelist[nummodes].modedesc, "%dx%dx%d",
					 modelist[nummodes].width,
					 modelist[nummodes].height,
					 modelist[nummodes].bpp);
		}
	}
}


/*
=================
VID_InitFullDIB
=================
*/
void VID_InitFullDIB (void)
{
	SDL_PixelFormat	format;
	SDL_Rect	**modes;
	Uint32		flags;
	int		i, j, modenum, originalnummodes;
	int		numlowresmodes;
	int		bpps[3] = {16, 32, 0}; // enumerate >8 bpp modes (no such thing as a 24-bit mode)

	originalnummodes = nummodes;
	modenum = 0;
	format.palette = NULL;

	// enumerate fullscreen modes
	// added SDL_HWSURFACE: http://sdl.beuc.net/sdl.wiki/SDL_SetVideoMode
	flags = SDL_DEFAULT_FLAGS | SDL_FULLSCREEN | SDL_HWSURFACE;

	for (i = 0;; i++)
	{
		if (!bpps[i]) break;
		if (nummodes >= MAX_MODE_LIST) break;

		memset (&format, 0, sizeof (SDL_PixelFormat));
		format.BitsPerPixel = bpps[i];
		modes = SDL_ListModes (&format, flags);

		if (modes == (SDL_Rect **) 0 || modes == (SDL_Rect **) - 1)
			continue;

		for (j = 0; modes[j]; j++)
		{
			if (modes[j]->w > MAXWIDTH || modes[j]->h > MAXHEIGHT || nummodes >= MAX_MODE_LIST) continue;

			// prevent intel rotated display modes from showing
			if (modes[j]->h > modes[j]->w) continue;

			VID_AddMode (modes[j]->w, modes[j]->h, bpps[i], false);

			if (!VID_DoesModeExist (originalnummodes, nummodes, modelist[nummodes].width, modelist[nummodes].height, modelist[nummodes].bpp))
			{
				nummodes++;
			}
		}

		modenum++;
	}

	if (nummodes == originalnummodes)
		Con_SafePrintf ("No fullscreen DIB modes found\n");

	// add any low-res modes that might not have been reported
	numlowresmodes = sizeof (lowresmodes) / sizeof (lowresmodes[0]);

	for (i = 0;; i++)
	{
		if (!bpps[i]) break;

		for (j = 0; j < numlowresmodes; j++)
		{
			if (nummodes >= MAX_MODE_LIST) continue;

			VID_AddMode (lowresmodes[j].width, lowresmodes[j].height, bpps[i], true);

			if (!VID_DoesModeExist (originalnummodes, nummodes, modelist[nummodes].width, modelist[nummodes].height, modelist[nummodes].bpp))
			{
				nummodes++;
			}
		}
	}

	// http://sdl.beuc.net/sdl.wiki/SDL_ListModes - "The array is not guaranteed to be sorted in any particular order"
	if (modelist[0].type == MODE_WINDOWED)
	{
		// modelist[0] is windowed (fixme - is this always true coming in here?)
		qsort (&modelist[1], nummodes - 1, sizeof (vmode_t), (sortfunc_t) VID_ModeSortFunc2);
	}
	else
	{
		// modelist[0] is fullscreen
		qsort (modelist, nummodes, sizeof (vmode_t), (sortfunc_t) VID_ModeSortFunc2);
	}
}

/*
===================
VID_Init
===================
*/
void VID_Init (void)
{
	const SDL_VideoInfo *info;
	int		i, existingmode;
	int		basenummodes, width, height, bpp, findbpp, done;

	// johnfitz -- clean up init readouts
	// Con_Printf("------------- Init Video -------------\n");
	// Con_Printf("%cVideo Init\n", 2);
	// johnfitz

	Cvar_RegisterVariable (&vid_fullscreen, VID_Changed_f); // johnfitz
	Cvar_RegisterVariable (&vid_width, VID_Changed_f); // johnfitz
	Cvar_RegisterVariable (&vid_height, VID_Changed_f); // johnfitz
	Cvar_RegisterVariable (&vid_bpp, VID_Changed_f); // johnfitz
	// Cvar_RegisterVariable (&vid_refreshrate, NULL); // johnfitz
	Cvar_RegisterVariable (&_windowed_mouse, NULL);

	// moved here from GL_Init as that now gets called on a restart
	Cmd_AddCommand ("gl_info", GL_Info_f); // johnfitz
	Cvar_RegisterVariable (&vid_vsync, VID_Changed_f); // johnfitz

	Cmd_AddCommand ("vid_unlock", VID_Unlock); // johnfitz
	Cmd_AddCommand ("vid_restart", VID_Restart_f); // johnfitz
	Cmd_AddCommand ("vid_forcerestart", VID_ForceRestart_f); // mh
	Cmd_AddCommand ("vid_test", VID_Test); // johnfitz
	Cmd_AddCommand ("vid_describecurrentmode", VID_DescribeCurrentMode_f);
	Cmd_AddCommand ("vid_describemodes", VID_DescribeModes_f);

	if (SDL_InitSubSystem (SDL_INIT_VIDEO) == -1)
		Sys_Error ("Could not initialize SDL Video");

	SDL_putenv ("SDL_VIDEO_CENTERED=center");

	VID_InitDIB ();
	basenummodes = nummodes = 1;
	VID_InitFullDIB ();

	// Config file is not read yet, so we don't know vid_fullscreen.value
	// Changed this to default to -window as otherwise it occasionally forces
	// two switches of video mode (window->fullscreen->window) which is bad S.A
	// It's still not perfect but, hell, this ancient code can be a pain
	if (!COM_CheckParm ("-fullscreen") && !COM_CheckParm ("-f"))
	{
		windowed = true;
		vid_default = MODE_WINDOWED;
	}
	else
	{
		Cvar_Set ("vid_fullscreen", "1");

		if (nummodes == 1)
			Sys_Error ("No RGB fullscreen modes available");

		windowed = false;

		if (COM_CheckParm ("-mode"))
		{
			vid_default = Q_atoi (com_argv[COM_CheckParm ("-mode") +1]);
		}
		else
		{
			if (COM_CheckParm ("-current"))
			{
				info = SDL_GetVideoInfo ();
				modelist[MODE_FULLSCREEN_DEFAULT].width = info->current_w;
				modelist[MODE_FULLSCREEN_DEFAULT].height = info->current_h;
				vid_default = MODE_FULLSCREEN_DEFAULT;
				leavecurrentmode = 1;
			}
			else
			{
				if (COM_CheckParm ("-width"))
				{
					width = Q_atoi (com_argv[COM_CheckParm ("-width") +1]);
				}
				else
				{
					width = 640;
				}

				if (COM_CheckParm ("-bpp"))
				{
					bpp = Q_atoi (com_argv[COM_CheckParm ("-bpp") +1]);
					findbpp = 0;
				}
				else
				{
					bpp = 15;
					findbpp = 1;
				}

				if (COM_CheckParm ("-height"))
					height = Q_atoi (com_argv[COM_CheckParm ("-height") +1]);
				else
					height = width * 3 / 4; // assume 4:3 aspect ratio

				// if they want to force it, add the specified mode to the list
				if (COM_CheckParm ("-force") && (nummodes < MAX_MODE_LIST))
				{
					modelist[nummodes].type = MODE_FULLSCREEN_DEFAULT;
					modelist[nummodes].width = width;
					modelist[nummodes].height = height;
					modelist[nummodes].modenum = 0;
					modelist[nummodes].halfscreen = 0;
					modelist[nummodes].dib = 1;
					modelist[nummodes].fullscreen = 1;
					modelist[nummodes].bpp = bpp;
					sprintf (modelist[nummodes].modedesc, "%dx%dx%d",
							 modelist[nummodes].width,
							 modelist[nummodes].height,
							 modelist[nummodes].bpp);

					for (i = nummodes, existingmode = 0; i < nummodes; i++)
					{
						if ((modelist[nummodes].width == modelist[i].width)   &&
								(modelist[nummodes].height == modelist[i].height) &&
								(modelist[nummodes].bpp == modelist[i].bpp))
						{
							existingmode = 1;
							break;
						}
					}

					if (!existingmode)
					{
						nummodes++;
					}
				}

				done = 0;

				do
				{
					if (COM_CheckParm ("-height"))
					{
						height = Q_atoi (com_argv[COM_CheckParm ("-height") +1]);

						for (i = 1, vid_default = 0; i < nummodes; i++)
						{
							if ((modelist[i].width == width) &&
									(modelist[i].height == height) &&
									(modelist[i].bpp == bpp))
							{
								vid_default = i;
								done = 1;
								break;
							}
						}
					}
					else
					{
						for (i = 1, vid_default = 0; i < nummodes; i++)
						{
							if ((modelist[i].width == width) && (modelist[i].bpp == bpp))
							{
								vid_default = i;
								done = 1;
								break;
							}
						}
					}

					if (!done)
					{
						if (findbpp)
						{
							switch (bpp)
							{
							case 15:
								bpp = 16;
								break;
							case 16:
								bpp = 32;
								break;
							case 32:
								bpp = 24;
								break;
							case 24:
								done = 1;
								break;
							}
						}
						else
						{
							done = 1;
						}
					}
				}
				while (!done);

				if (!vid_default)
				{
					Sys_Error ("Specified video mode not available");
				}
			}
		}
	}

	vid_initialized = true;

	vid.maxwarpwidth = WARP_WIDTH;
	vid.maxwarpheight = WARP_HEIGHT;
	vid.colormap = host_colormap;
	vid.fullbright = 256 - LittleLong (*((int *) vid.colormap + 2048));

	// set window icon
	PL_SetWindowIcon ();

#if 0
	// set up the desired pixel format attributes for OpenGL
	// this currently segfaults on Linux if we explicitly request no accumulation buffer - SHIT!
	SDL_GL_SetAttribute (SDL_GL_RED_SIZE, 8);
	SDL_GL_SetAttribute (SDL_GL_GREEN_SIZE, 8);
	SDL_GL_SetAttribute (SDL_GL_BLUE_SIZE, 8);
	SDL_GL_SetAttribute (SDL_GL_ALPHA_SIZE, 8);
	SDL_GL_SetAttribute (SDL_GL_BUFFER_SIZE, 32);
	SDL_GL_SetAttribute (SDL_GL_DOUBLEBUFFER, 1);
	SDL_GL_SetAttribute (SDL_GL_DEPTH_SIZE, 32);
	SDL_GL_SetAttribute (SDL_GL_STENCIL_SIZE, 0);
	SDL_GL_SetAttribute (SDL_GL_ACCUM_RED_SIZE, 0);
	SDL_GL_SetAttribute (SDL_GL_ACCUM_GREEN_SIZE, 0);
	SDL_GL_SetAttribute (SDL_GL_ACCUM_BLUE_SIZE, 0);
	SDL_GL_SetAttribute (SDL_GL_ACCUM_ALPHA_SIZE, 0);
	SDL_GL_SetAttribute (SDL_GL_STEREO, 0);
	SDL_GL_SetAttribute (SDL_GL_MULTISAMPLEBUFFERS, 0);
	SDL_GL_SetAttribute (SDL_GL_MULTISAMPLESAMPLES, 0);
	SDL_GL_SetAttribute (SDL_GL_ACCELERATED_VISUAL, 1);
	SDL_GL_SetAttribute (SDL_GL_SWAP_CONTROL, 0);
#endif

	// seems that OpenGL (or at least SDL) gives us an accumulation buffer whether we want one or not
	// which is kinda crap
	SDL_GL_SetAttribute (SDL_GL_RED_SIZE, 8);
	SDL_GL_SetAttribute (SDL_GL_GREEN_SIZE, 8);
	SDL_GL_SetAttribute (SDL_GL_BLUE_SIZE, 8);
	SDL_GL_SetAttribute (SDL_GL_ALPHA_SIZE, 8);
	SDL_GL_SetAttribute (SDL_GL_BUFFER_SIZE, 32);
	SDL_GL_SetAttribute (SDL_GL_DOUBLEBUFFER, 1);

	// prefer D24S8 so that we can do shadows properly
	SDL_GL_SetAttribute (SDL_GL_DEPTH_SIZE, 24);
	SDL_GL_SetAttribute (SDL_GL_STENCIL_SIZE, 8);

	VID_SetMode (vid_default);

	// johnfitz -- removed code creating "glquake" subdirectory

	vid_realmode = vid_modenum;

	vid_menucmdfn = VID_Menu_f; // johnfitz
	vid_menudrawfn = VID_MenuDraw;
	vid_menukeyfn = VID_MenuKey;

	strcpy (badmode.modedesc, "Bad mode");
	vid_canalttab = true;

	VID_Gamma_Init (); // johnfitz
	VID_Menu_Init (); // johnfitz

	// johnfitz -- command line vid settings should override config file settings.
	// so we have to lock the vid mode from now until after all config files are read.
	if (COM_CheckParm ("-width") || COM_CheckParm ("-height") ||
			COM_CheckParm ("-bpp")				||
			COM_CheckParm ("-window") || COM_CheckParm ("-w") ||
			COM_CheckParm ("-fullscreen") || COM_CheckParm ("-f"))
	{
		vid_locked = true;
	}

	// johnfitz

	// The problem here is (say) previous video mode is 1024x768 windowed
	// And we call "fitzquake -w". This disables setting of 1024x768, and when video lock
	// is removed, we get 800x600.
}


/*
================
VID_SyncCvars -- johnfitz -- set vid cvars to match current video mode
================
*/
void VID_SyncCvars (void)
{
	int swap_control;

	Cvar_Set ("vid_width", va ("%i", modelist[vid_default].width));
	Cvar_Set ("vid_height", va ("%i", modelist[vid_default].height));
	Cvar_Set ("vid_bpp", va ("%i", modelist[vid_default].bpp));
	Cvar_Set ("vid_fullscreen", (windowed) ? "0" : "1");

	if (SDL_GL_GetAttribute (SDL_GL_SWAP_CONTROL, &swap_control) == 0)
		Cvar_Set ("vid_vsync", (swap_control) ? "1" : "0");
	else Cvar_Set ("vid_vsync", 0);

	Cvar_Set ("vid_vsync", 0);
}


// ==========================================================================
//
//  NEW VIDEO MENU -- johnfitz
//
// ==========================================================================

extern void M_Menu_Options_f (void);
extern void M_Print (int cx, int cy, char *str);
extern void M_PrintWhite (int cx, int cy, char *str);
extern void M_DrawCharacter (int cx, int line, int num);
extern void M_DrawTransPic (int x, int y, qpic_t *pic);
extern void M_DrawPic (int x, int y, qpic_t *pic);
extern void M_DrawCheckbox (int x, int y, int on);
int M_DrawBanner (char *bannerpic);
void M_DrawQPlaque (void);

extern qboolean	m_entersound;

#define VIDEO_OPTIONS_ITEMS 8
int		video_cursor_table[] = {32, 44, 52, 60, 68, 80, 92, 100};
int		video_options_cursor = 0;

typedef struct {int width, height;} vid_menu_mode;

// TODO: replace these fixed-length arrays with hunk_allocated buffers

vid_menu_mode vid_menu_modes[MAX_MODE_LIST];
int vid_menu_nummodes = 0;

int vid_menu_bpps[4];
int vid_menu_numbpps = 0;

int vid_menu_rates[20];
int vid_menu_numrates = 0;

extern cvar_t scr_widescreenfov;

void M_CenterPrint (int cy, char *str);
void M_CenterPrintWhite (int cy, char *str);

/*
================
VID_Menu_Init
================
*/
void VID_Menu_Init (void)
{
	int i, j, h, w;

	for (i = 1; i < nummodes; i++) // start i at mode 1 because 0 is windowed mode
	{
		w = modelist[i].width;
		h = modelist[i].height;

		for (j = 0; j < vid_menu_nummodes; j++)
		{
			if (vid_menu_modes[j].width == w &&
					vid_menu_modes[j].height == h)
				break;
		}

		if (j == vid_menu_nummodes)
		{
			vid_menu_modes[j].width = w;
			vid_menu_modes[j].height = h;
			vid_menu_nummodes++;
		}
	}
}

/*
================
VID_Menu_RebuildBppList

regenerates bpp list based on current vid_width and vid_height
================
*/
void VID_Menu_RebuildBppList (void)
{
	int i, j, b;

	vid_menu_numbpps = 0;

	for (i = 1; i < nummodes; i++) // start i at mode 1 because 0 is windowed mode
	{
		// bpp list is limited to bpps available with current width/height
		if (modelist[i].width != vid_width.value ||
				modelist[i].height != vid_height.value)
			continue;

		b = modelist[i].bpp;

		for (j = 0; j < vid_menu_numbpps; j++)
		{
			if (vid_menu_bpps[j] == b)
				break;
		}

		if (j == vid_menu_numbpps)
		{
			vid_menu_bpps[j] = b;
			vid_menu_numbpps++;
		}
	}

	// if there are no valid fullscreen bpps for this width/height, just pick one
	if (vid_menu_numbpps == 0)
	{
		Cvar_SetValue ("vid_bpp", (float) modelist[0].bpp);
		return;
	}

	// if vid_bpp is not in the new list, change vid_bpp
	for (i = 0; i < vid_menu_numbpps; i++)
		if (vid_menu_bpps[i] == (int) (vid_bpp.value))
			break;

	if (i == vid_menu_numbpps)
		Cvar_SetValue ("vid_bpp", (float) vid_menu_bpps[0]);
}

/*
================
VID_Menu_ChooseNextMode

chooses next resolution in order, then updates vid_width and
vid_height cvars, then updates bpp and refreshrate lists
================
*/
void VID_Menu_ChooseNextMode (int dir)
{
	int i;

	for (i = 0; i < vid_menu_nummodes; i++)
	{
		if (vid_menu_modes[i].width == vid_width.value &&
			vid_menu_modes[i].height == vid_height.value)
			break;
	}

	if (i == vid_menu_nummodes) // can't find it in list, so it must be a custom windowed res
	{
		i = 0;
	}
	else
	{
		i += dir;

		if (i >= vid_menu_nummodes)
			i = 0;
		else if (i < 0)
			i = vid_menu_nummodes - 1;
	}

	Cvar_SetValue ("vid_width", (float) vid_menu_modes[i].width);
	Cvar_SetValue ("vid_height", (float) vid_menu_modes[i].height);
	VID_Menu_RebuildBppList ();
}

/*
================
VID_Menu_ChooseNextBpp

chooses next bpp in order, then updates vid_bpp cvar, then updates refreshrate list
================
*/
void VID_Menu_ChooseNextBpp (int dir)
{
	int i;

	for (i = 0; i < vid_menu_numbpps; i++)
	{
		if (vid_menu_bpps[i] == vid_bpp.value)
			break;
	}

	if (i == vid_menu_numbpps) // can't find it in list
	{
		i = 0;
	}
	else
	{
		i += dir;

		if (i >= vid_menu_numbpps)
			i = 0;
		else if (i < 0)
			i = vid_menu_numbpps - 1;
	}

	Cvar_SetValue ("vid_bpp", (float) vid_menu_bpps[i]);
}

/*
================
VID_Menu_ChooseNextRate

chooses next refresh rate in order, then updates vid_refreshrate cvar
================
*/
void VID_Menu_ChooseNextRate (int dir)
{
	int i;

	for (i = 0; i < vid_menu_numrates; i++)
	{
		if (vid_menu_rates[i] == vid_refreshrate.value)
			break;
	}

	if (i == vid_menu_numrates) // can't find it in list
	{
		i = 0;
	}
	else
	{
		i += dir;

		if (i >= vid_menu_numrates)
			i = 0;
		else if (i < 0)
			i = vid_menu_numrates - 1;
	}

	Cvar_SetValue ("vid_refreshrate", (float) vid_menu_rates[i]);
}

/*
================
VID_MenuKey
================
*/
void VID_MenuKey (int key)
{
	switch (key)
	{
	case K_ESCAPE:
		VID_SyncCvars (); // sync cvars before leaving menu. FIXME: there are other ways to leave menu
		S_LocalSound ("misc/menu1.wav");
		M_Menu_Options_f ();
		break;

	case K_UPARROW:
		S_LocalSound ("misc/menu1.wav");
		video_options_cursor--;

		if (video_options_cursor < 0)
			video_options_cursor = VIDEO_OPTIONS_ITEMS - 1;

		break;

	case K_DOWNARROW:
		S_LocalSound ("misc/menu1.wav");
		video_options_cursor++;

		if (video_options_cursor >= VIDEO_OPTIONS_ITEMS)
			video_options_cursor = 0;

		break;

	case K_LEFTARROW:
		S_LocalSound ("misc/menu3.wav");

		switch (video_options_cursor)
		{
		case 0:
			VID_Menu_ChooseNextMode (-1);
			break;
		case 1:
			VID_Menu_ChooseNextBpp (-1);
			break;
		case 2:
			VID_Menu_ChooseNextRate (-1);
			break;
		case 3:
			Cbuf_AddText ("toggle vid_fullscreen\n");
			break;
		case 4:
			Cbuf_AddText ("toggle vid_vsync\n"); // kristian
			break;
		case 5:
			Cvar_SetValue ("scr_widescreenfov", !scr_widescreenfov.value);
			break;
		case 6:
		case 7:
		default:
			break;
		}

		break;

	case K_RIGHTARROW:
		S_LocalSound ("misc/menu3.wav");

		switch (video_options_cursor)
		{
		case 0:
			VID_Menu_ChooseNextMode (1);
			break;
		case 1:
			VID_Menu_ChooseNextBpp (1);
			break;
		case 2:
			VID_Menu_ChooseNextRate (1);
			break;
		case 3:
			Cbuf_AddText ("toggle vid_fullscreen\n");
			break;
		case 4:
			Cbuf_AddText ("toggle vid_vsync\n");
			break;
		case 5:
			Cvar_SetValue ("scr_widescreenfov", !scr_widescreenfov.value);
			break;
		case 6:
		case 7:
		default:
			break;
		}

		break;

	case K_ENTER:
		m_entersound = true;

		switch (video_options_cursor)
		{
		case 0:
			VID_Menu_ChooseNextMode (1);
			break;
		case 1:
			VID_Menu_ChooseNextBpp (1);
			break;
		case 2:
			VID_Menu_ChooseNextRate (1);
			break;
		case 3:
			Cbuf_AddText ("toggle vid_fullscreen\n");
			break;
		case 4:
			Cbuf_AddText ("toggle vid_vsync\n");
			break;
		case 5:
			Cvar_SetValue ("scr_widescreenfov", !scr_widescreenfov.value);
			break;
		case 6:
			Cbuf_AddText ("vid_test\n");
			break;
		case 7:
			Cbuf_AddText ("vid_restart\n");
			key_dest = key_game;
			m_state = m_none;
			IN_Activate ();
			break;
		default:
			break;
		}

		break;

	default:
		break;
	}
}


qboolean VID_MenuWarn (int width, int height, int bpp)
{
	int i;

	for (i = 0; i < nummodes; i++)
	{
		if (modelist[i].width == width && modelist[i].height == height && modelist[i].bpp == bpp)
			return modelist[i].warn;
	}

	return false;
}


/*
================
VID_MenuDraw
================
*/
void VID_MenuDraw (void)
{
	int i = 0;

	M_DrawQPlaque ();
	M_DrawBanner ("gfx/vidmodes.lmp");

	// title
	// title = "Video Options";
	// M_PrintWhite ((320 - 8 * strlen (title)) / 2, 32, title);

	// options
	M_Print (16, video_cursor_table[i], "        Video mode");
	M_PrintWhite (184, video_cursor_table[i], va ("%ix%i", (int) vid_width.value, (int) vid_height.value));
	i++;

	M_Print (16, video_cursor_table[i], "       Color depth");
	M_PrintWhite (184, video_cursor_table[i], va ("%i", (int) vid_bpp.value));
	i++;

	M_Print (16, video_cursor_table[i], "      Refresh rate");
	// 	M_PrintWhite (184, video_cursor_table[i], va("%i Hz", (int)vid_refreshrate.value)); refresh rates are disabled for now -- kristian
	M_PrintWhite (184, video_cursor_table[i], "N/A");
	i++;

	M_Print (16, video_cursor_table[i], "        Fullscreen");
	M_DrawCheckbox (184, video_cursor_table[i], (int) vid_fullscreen.value);
	i++;

	// added vsync to the video menu -- kristian
	M_Print (16, video_cursor_table[i], "     Vertical Sync");

	if (gl_swap_control)
		M_DrawCheckbox (184, video_cursor_table[i], (int) vid_vsync.value);
	else M_PrintWhite (184, video_cursor_table[i], "N/A");

	i++;

	M_Print (16, video_cursor_table[i], "    Widescreen FOV");
	M_DrawCheckbox (184, video_cursor_table[i], (int) scr_widescreenfov.value);
	i++;

	M_Print (16, video_cursor_table[i], "      Test changes");
	i++;

	M_Print (16, video_cursor_table[i], "     Apply changes");

	// cursor
	M_DrawCharacter (168, video_cursor_table[video_options_cursor], 12 + ((int) (realtime * 4) & 1));

	// warn if we're trying to select a fullscreen low-res mode
	if (VID_MenuWarn ((int) vid_width.value, (int) vid_height.value, (int) vid_bpp.value) && vid_fullscreen.value)
	{
		int y = video_cursor_table[i] + 24;
		char *line = "\35\36\36\36\36\36\36\36\36\36\36\36\36\37";

		M_CenterPrintWhite (y, "Warning");
		M_CenterPrint (y + 12, line);
		M_CenterPrint (y + 24, "This mode may not be");
		M_CenterPrint (y + 32, "supported by your hardware!");
		M_CenterPrint (y + 44, "Use at your own risk");
	}

	// notes          "345678901234567890123456789012345678"
	// 	M_Print (16, 172, "Windowed modes always use the desk- ");
	// 	M_Print (16, 180, "top color depth, and can never be   ");
	// 	M_Print (16, 188, "larger than the desktop resolution. ");
}

/*
================
VID_Menu_f
================
*/
void VID_Menu_f (void)
{
	key_dest = key_menu;
	m_state = m_video;
	m_entersound = true;

	// set all the cvars to match the current mode when entering the menu
	VID_SyncCvars ();

	// set up bpp and rate lists based on current cvars
	VID_Menu_RebuildBppList ();
}
