/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef _QUAKE_PROGS_H
#define _QUAKE_PROGS_H

#include "pr_comp.h"			// defs shared with qcc
#include "progdefs.h"			// generated by program cdefs

typedef union eval_s
{
	string_t		string;
	float			_float;
	float			vector[3];
	func_t			function;
	int				_int;
	int				edict;
} eval_t;


typedef struct sv_efrag_s
{
	int leafnum;
	struct sv_efrag_s *next;
} sv_efrag_t;

typedef struct edict_s
{
	qboolean		free;
	link_t			area;				// linked to a division node or leaf
	entity_state_t	baseline;
	sv_efrag_t		*visleafs;
	int				visframe;
	unsigned char	alpha;				// johnfitz -- hack to support alpha since it's not part of entvars_t
	unsigned char	scale;				// mh -- hack to support scale since it's not part of entvars_t
	qboolean		sendinterval;		// johnfitz -- send time until nextthink to client for better lerp timing
	float			freetime;			// pr_progs->time when the object was freed

	int				ednum;
	int				prog;

	// LordHavoc: for MOVETYPE_STEP interpolation
	vec3_t		steporigin;
	vec3_t		stepangles;
	vec3_t		stepoldorigin;
	vec3_t		stepoldangles;
	float		steplerptime;

	// fixme - different for csprogs????
	entvars_t		v;					// C exported fields from progs
	// other fields from progs come immediately after
} edict_t;

void SV_ReturnEfrags (edict_t *ed);

void ED_Memset0 (edict_t *e);

#define	EDICT_FROM_AREA(l) STRUCT_FROM_LINK(l,edict_t,area)

extern int eval_drawonlytoclient; // Team Xlink DP_SV_DRAWONLYTOCLIENT
extern int eval_nodrawtoclient; // Team Xlink DP_SV_NODRAWTOCLIENT
extern int eval_alpha;
extern int eval_scale;

#define GETEDICTFIELDVALUE(ed, fieldoffset) (fieldoffset ? (eval_t*)((char*)&ed->v + fieldoffset) : NULL)

typedef void (*builtin_t) (void);

// 2001-09-14 Enhanced BuiltIn Function System (EBFS) by Maddes  start
typedef struct ebfs_builtin_s
{
	int			default_funcno;
	char		*funcname;
	builtin_t	function;
	int			funcno;
} ebfs_builtin_t;

typedef struct
{
	int				s;
	dfunction_t		*f;
	qboolean		nostack;
} prstack_t;

#define	MAX_STACK_DEPTH		32
#define	LOCALSTACK_SIZE		2048

#define	MAX_FIELD_LEN	64
#define GEFV_CACHESIZE	16	// bumped this because there are gonna be loads more fields we wanna cache

typedef struct
{
	ddef_t	*pcache;
	char	field[MAX_FIELD_LEN];
} gefv_cache;

typedef struct prvm_s
{
	dprograms_t *progs;

	gefv_cache GEFV_Cache[GEFV_CACHESIZE];

	// fixme - could these be a union???
	float *globals;
	globalvars_t *global_struct;

	dfunction_t *functions;
	dstatement_t *statements;

	ddef_t *fielddefs;
	ddef_t *globaldefs;

	int edict_size;	// in bytes
	int edict_shift;

	qboolean trace;
	dfunction_t *xfunction;
	int xstatement;

	unsigned short crc;

	builtin_t *builtins;
	int numbuiltins;

	ebfs_builtin_t *ebfs_builtins;
	int ebfs_numbuiltins;

	prstack_t stack[MAX_STACK_DEPTH];
	int stackdepth;

	int localstack[LOCALSTACK_SIZE];
	int localstack_used;

	int prargc;

	char *strings;
	int stringsize;

	char **knownstrings;
	int numknownstrings;
	int maxknownstrings;

	int opprofiles[256];

	double time;		// moved here from sv because it's going to be different on the client (fixme - also move from client?  what if there is no csprogs?)

	int first_edict;	// that isn't a client
	int num_edicts;
	int max_edicts;

	edict_t *edicts[MAX_EDICTS];
} prvm_t;

extern prvm_t cl_progs;
extern prvm_t sv_progs;		// server-side vm
extern prvm_t *pr_progs;	// currently active vm

// ============================================================================

void PR_Init (void);

void PR_ExecuteProgram (prvm_t *progs, func_t fnum);
qboolean PR_LoadProgs (prvm_t *progs, char *filename, qboolean mandatory, int crc);

void PR_Profile_f (void);

char *PR_GetString (int num);
int PR_SetString (char *s);

edict_t *ED_Alloc (void);
void ED_Free (edict_t *ed);

void ED_Print (edict_t *ed);
void ED_Write (FILE *f, edict_t *ed);
char *ED_ParseEdict (char *data, edict_t *ent);

void ED_WriteGlobals (FILE *f);
void ED_ParseGlobals (char *data);

void ED_LoadFromFile (char *data);

edict_t *GetEdictForNumber (int n);
int GetNumberForEdict (edict_t *e);

#define	NEXT_EDICT(e) (pr_progs->edicts[(e)->ednum + 1])

#define	EDICT_TO_PROG(e) ((e)->prog)
#define PROG_TO_EDICT(e) (pr_progs->edicts[e >> pr_progs->edict_shift])

// ============================================================================

// this fixes a load of crap with NaNs in the globals table
#define	G_FLOAT(o) (pr_progs->globals[o])
#define	G_INT(o) (((int *) pr_progs->globals)[o])
#define	G_EDICT(o) (PROG_TO_EDICT(((int *) pr_progs->globals)[o]))
#define G_EDICTNUM(o) GetNumberForEdict (G_EDICT(o))
#define	G_VECTOR(o) (&pr_progs->globals[o])
#define	G_STRING(o) (PR_GetString (((int *) pr_progs->globals)[o]))
#define	G_FUNCTION(o) (((func_t *) pr_progs->globals)[o])

#define	E_FLOAT(e,o) (((float*)&e->v)[o])
#define	E_INT(e,o) (*(int *)&((float*)&e->v)[o])
#define	E_VECTOR(e,o) (&((float*)&e->v)[o])
#define	E_STRING(e,o) (PR_GetString(*(string_t *)&((float*)&e->v)[o]))

extern	int		type_size[8];

#define PR_DEFAULT_FUNCNO_EXTENSION_FIND	99	// 2001-10-20 Extension System by LordHavoc
#define PR_DEFAULT_FUNCNO_BUILTIN_FIND	100

// 2001-10-20 Extension System by LordHavoc  start
extern char *pr_extensions[];
// 2001-10-20 Extension System by LordHavoc  end

extern cvar_t	pr_builtin_find;
extern cvar_t	pr_builtin_remap;

#define PR_DEFAULT_FUNCNO_EXTENSION_FIND	99	// 2001-10-20 Extension System by Lord Havoc/Maddes
// 2001-09-14 Enhanced BuiltIn Function System (EBFS) by Maddes  end

void PR_RunError (const char *error, ...) __attribute__ ((__format__ (__printf__, 1, 2)));

void ED_PrintEdicts (void);
void ED_PrintNum (int ent);

dfunction_t *ED_FindFunction (char *name);
eval_t *GetEdictFieldValue (edict_t *ed, char *field);

void PR_SetActiveProgs (prvm_t *progs);
void PR_InitEBFS (void);

#endif	/* _QUAKE_PROGS_H */

