/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others
Copyright (C) 2007-2008 Kristian Duske

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
// r_sprite.c -- sprite model rendering

#include "quakedef.h"


// vecc = veca + scale * vecb (wrong way round again!!!!!)
/*
	// veca, scale, vecb, vecc
	VectorMA (origin + f1 * upvec, point);
	VectorMA (point + f2 * rightvec, point);
	VectorScale (point, scale, point);
*/

#define sprite_vp \
	"!!ARBvp1.0\n" \
	"TEMP point, scaled;\n" \
	"MUL scaled, vertex.attrib[4], vertex.attrib[4].z;\n" \
	"MAD point, vertex.attrib[2], scaled.x, vertex.attrib[0];\n" \
	"MAD point, vertex.attrib[3], scaled.y, point;\n" \
	"MOV point.w, 1; # help ma boab!\n" \
	TRANSFORMVERTEX ("result.position", "point") \
	"MOV result.texcoord[0], vertex.attrib[5];\n" \
	"MOV result.color, vertex.attrib[1];\n" \
	"END\n"

#define sprite_fp \
	"!!ARBfp1.0\n" \
	"TEMP col1;\n" \
	"TEX col1, fragment.texcoord[0], texture[0], 2D;\n" \
	"MUL result.color, col1, fragment.color;\n" \
	"END\n"

GLuint arb_sprite_vp = 0;
GLuint arb_sprite_fp = 0;
GLuint arb_sprite_fp_fog = 0;


void Sprite_CreateShaders (void)
{
	arb_sprite_vp = GL_CreateProgram (GL_VERTEX_PROGRAM_ARB, sprite_vp, false);
	arb_sprite_fp = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, sprite_fp, false);
	arb_sprite_fp_fog = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, sprite_fp, true);
}


typedef struct spritevert_s
{
	float origin[3];
	float upvec[3];
	float rightvec[3];

	// scale must follow framevecs as this is a 3-component texture in the vertex array pointer
	float framevecs[2];
	float scale;

	union
	{
		byte color[4];
		unsigned int rgba;
	};

	float st[2];
} spritevert_t;


GLuint r_lastspritetexture = 0;
spritevert_t *r_spriteverts = NULL;
int r_numsprites = 0;

extern unsigned int r_parttexcoordbuffer;

void R_InvalidateSprite (void)
{
	// used when we want to force a state update
	r_lastspritetexture = 0;
}


void R_SpriteBegin (void)
{
	glLoadMatrixf (r_world_matrix.m16);

	r_spriteverts = (spritevert_t *) scratchbuf;
	r_numsprites = 0;

	GL_SetVertexProgram (arb_sprite_vp);

	if (Fog_GetDensity () > 0)
		GL_SetFragmentProgram (arb_sprite_fp_fog);
	else GL_SetFragmentProgram (arb_sprite_fp);

	GL_EnableVertexArrays (VAA_0 | VAA_1 | VAA_2 | VAA_3 | VAA_4 | VAA_5);
	GL_VertexArrayPointer (0, 0, 3, GL_FLOAT, GL_FALSE, sizeof (spritevert_t), r_spriteverts->origin);
	GL_VertexArrayPointer (0, 1, 4, GL_UNSIGNED_BYTE, GL_TRUE, sizeof (spritevert_t), r_spriteverts->color);
	GL_VertexArrayPointer (0, 2, 3, GL_FLOAT, GL_FALSE, sizeof (spritevert_t), r_spriteverts->upvec);
	GL_VertexArrayPointer (0, 3, 3, GL_FLOAT, GL_FALSE, sizeof (spritevert_t), r_spriteverts->rightvec);
	GL_VertexArrayPointer (0, 4, 3, GL_FLOAT, GL_FALSE, sizeof (spritevert_t), r_spriteverts->framevecs);
	GL_VertexArrayPointer (0, 5, 2, GL_FLOAT, GL_FALSE, sizeof (spritevert_t), r_spriteverts->st);
}


void R_SpriteEnd (void)
{
	if (r_numsprites)
	{
		GL_DrawArrays (GL_QUADS, 0, r_numsprites * 4);
		r_spriteverts = (spritevert_t *) scratchbuf;
		r_numsprites = 0;
	}
}


#define EmitSpriteVertex(vert, f1, f2, s, t) \
	vert.origin[0] = origin[0]; \
	vert.origin[1] = origin[1]; \
	vert.origin[2] = origin[2]; \
	vert.upvec[0] = up[0]; \
	vert.upvec[1] = up[1]; \
	vert.upvec[2] = up[2]; \
	vert.rightvec[0] = right[0]; \
	vert.rightvec[1] = right[1]; \
	vert.rightvec[2] = right[2]; \
	vert.framevecs[0] = f1; \
	vert.framevecs[1] = f2; \
	vert.scale = scale; \
	vert.color[0] = BYTE_CLAMPF (color[0]); \
	vert.color[1] = BYTE_CLAMPF (color[1]); \
	vert.color[2] = BYTE_CLAMPF (color[2]); \
	vert.color[3] = BYTE_CLAMPF (alpha); \
	vert.st[0] = s; \
	vert.st[1] = t;

void R_RenderSpritePolygon (mspriteframe_t *frame, float *origin, float *up, float *right, float *color, float alpha, float scale)
{
	EmitSpriteVertex (r_spriteverts[0], frame->up, frame->left, 0, 0);
	EmitSpriteVertex (r_spriteverts[1], frame->up, frame->right, frame->smax, 0);
	EmitSpriteVertex (r_spriteverts[2], frame->down, frame->right, frame->smax, frame->tmax);
	EmitSpriteVertex (r_spriteverts[3], frame->down, frame->left, 0, frame->tmax);

	r_spriteverts += 4;
	r_numsprites++;
}


/*
================
R_GetSpriteFrame
================
*/
mspriteframe_t *R_GetSpriteFrame (entity_t *e)
{
	msprite_t		*psprite;
	mspritegroup_t	*pspritegroup;
	mspriteframe_t	*pspriteframe;
	int				i, numframes, frame;
	float			*pintervals, fullinterval, targettime, time;

	psprite = (msprite_t *) e->model->cache.data;
	frame = e->frame;

	if ((frame >= psprite->numframes) || (frame < 0))
	{
		Con_DPrintf ("R_DrawSprite: no such frame %d\n", frame);
		frame = 0;
	}

	if (psprite->frames[frame].type == SPR_SINGLE)
	{
		pspriteframe = psprite->frames[frame].frameptr;
		// Con_Printf ("R_GetSpriteFrame returned %i\n", frame);
	}
	else
	{
		pspritegroup = (mspritegroup_t *) psprite->frames[frame].frameptr;
		pintervals = pspritegroup->intervals;
		numframes = pspritegroup->numframes;
		fullinterval = pintervals[numframes-1];

		time = cl.time + e->syncbase;

		// when loading in Mod_LoadSpriteGroup, we guaranteed all interval values
		// are positive, so we don't have to worry about division by 0
		targettime = time - ((int) (time / fullinterval)) * fullinterval;

		for (i = 0; i < (numframes - 1); i++)
		{
			if (pintervals[i] > targettime)
				break;
		}

		pspriteframe = pspritegroup->frames[i];
		// Con_Printf ("R_GetSpriteFrame returned %i[%i]\n", frame, i);
	}

	return pspriteframe;
}


/*
=================
R_DrawSpriteModel -- johnfitz -- rewritten: now supports all orientations
=================
*/
void R_DrawSpriteModel (entity_t *e)
{
	vec3_t			v_forward, v_right, v_up;
	msprite_t		*psprite;
	mspriteframe_t	*frame;
	float			*s_up, *s_right;
	float			angle, sr, cr;
	float			defaultcolor[3] = {1, 1, 1};
	float			*color = defaultcolor;

	frame = R_GetSpriteFrame (e);
	psprite = (msprite_t *) e->model->cache.data;

	if (psprite->litsprite)
	{
		vec3_t lightcolor;

		R_LightPoint (e, lightcolor);
		VectorScale (lightcolor, 1.0f / 255.0f, lightcolor);
		color = lightcolor;
	}

	switch (psprite->type)
	{
	case SPR_VP_PARALLEL_UPRIGHT: // faces view plane, up is towards the heavens
		v_up[0] = 0;
		v_up[1] = 0;
		v_up[2] = 1;
		s_up = v_up;
		s_right = vright;
		break;

	case SPR_FACING_UPRIGHT: // faces camera origin, up is towards the heavens
		VectorSubtract (e->origin, r_origin, v_forward);
		v_forward[2] = 0;
		VectorNormalizeFast (v_forward);
		v_right[0] = v_forward[1];
		v_right[1] = -v_forward[0];
		v_right[2] = 0;
		v_up[0] = 0;
		v_up[1] = 0;
		v_up[2] = 1;
		s_up = v_up;
		s_right = v_right;
		break;

	case SPR_VP_PARALLEL: // faces view plane, up is towards the top of the screen
		s_up = vup;
		s_right = vright;
		break;

	case SPR_ORIENTED: // pitch yaw roll are independent of camera
		AngleVectors (e->angles, v_forward, v_right, v_up);
		s_up = v_up;
		s_right = v_right;
		break;

	case SPR_VP_PARALLEL_ORIENTED: // faces view plane, but obeys roll value
		angle = e->angles[ROLL] * M_PI_DIV_180;
		sr = sin (angle);
		cr = cos (angle);
		v_right[0] = vright[0] * cr + vup[0] * sr;
		v_right[1] = vright[1] * cr + vup[1] * sr;
		v_right[2] = vright[2] * cr + vup[2] * sr;
		v_up[0] = vright[0] * -sr + vup[0] * cr;
		v_up[1] = vright[1] * -sr + vup[1] * cr;
		v_up[2] = vright[2] * -sr + vup[2] * cr;
		s_up = v_up;
		s_right = v_right;
		break;

	default:
		return;
	}

	// mh - polygon offset is crap because:
	// 1) depth buffers aren't the same on different cards
	// 2) 16-bit depth is different to 24-bit depth
	// 3) the depth buffer is non-linear so it gives different results depending on the depth
	// so in other words we don't use it

	// if it overflows invalidate the cached state to force a flush and reset
	if (r_numsprites == 4096)
		r_lastspritetexture = 0;

	if (frame->gltexture->texnum != r_lastspritetexture)
	{
		if (r_numsprites)
		{
			GL_DrawArrays (GL_QUADS, 0, r_numsprites * 4);
			r_spriteverts = (spritevert_t *) scratchbuf;
			r_numsprites = 0;
		}

		// only if it changes
		GL_BindTexture (GL_TEXTURE0, frame->gltexture);

		r_lastspritetexture = frame->gltexture->texnum;
	}

	// fucking fitz alpha
	if (e->alpha < 1)
		R_RenderSpritePolygon (frame, e->origin, s_up, s_right, color, 1.0f, (float) e->scale / 16.0f);
	else R_RenderSpritePolygon (frame, e->origin, s_up, s_right, color, (float) e->alpha / 255.0f, (float) e->scale / 16.0f);
}


void R_DrawSpriteModels (void)
{
	int i;

	if (!r_drawentities.value) return;

	// sprites just go onto the alpha list for later drawing
	for (i = 0; i < cl_numvisedicts; i++)
	{
		if (cl_visedicts[i]->model->type != mod_sprite) continue;

		// we can't sort by texture as sprites may be interleaved with different object types at different depths
		R_AddEntityToAlphaList (cl_visedicts[i]);
	}
}



